# -*- coding: utf-8 -*-

__version__ = "0.3.5"


import codecs
import decimal
from itertools import product
from string import ascii_letters

import click


# The arguments for -f and -s come in as raw strings, but we
# need to be able to interpret things like \t and \n as escape
# sequences, not literals.
def interpret(s):
    if s:
        return codecs.escape_decode(bytes(s, "utf8"))[0].decode("utf8")
    else:
        return None


class Counter(object):
    """
    Counter class to creat the range for a given counter. Either the "main" counter
    or an "--also" counter. The counter decides how to count (chars, ints or floats)
    based on the values of start, stop and step.
    """

    def __init__(self, start, stop, step, reverse=False):
        if str(start) in ascii_letters or str(stop) in ascii_letters:
            if start not in ascii_letters:
                raise ValueError(
                    f"For character counters 'START' needs to be a character. Given: {start}"
                )
            if stop not in ascii_letters:
                raise ValueError(
                    f"For character counters 'STOP' needs to be a character. Given: {stop}"
                )
            self.start = ord(start)
            self.stop = ord(stop)
            self.type = chr
        else:
            try:
                self.start = decimal.Decimal(start)
            except decimal.InvalidOperation:
                raise ValueError(
                    f"Invalid value for '[START]': '{start}' is not a valid number."
                )
            try:
                self.stop = decimal.Decimal(stop)
            except decimal.InvalidOperation:
                raise ValueError(
                    f"Invalid value for 'STOP': '{stop}' is not a valid number."
                )

            if "." in str(start) or "." in str(stop) or "." in str(step):
                self.type = float
            else:
                self.type = int

        try:
            self.step = abs(decimal.Decimal(step))
        except decimal.InvalidOperation:
            raise ValueError(
                f"Invalid value for '[STEP]': '{step}' is not a valid number."
            )

        self.reversed = False  # reverse
        if self.start > self.stop:
            # self.start, self.stop = self.stop, self.start
            self.step = -1 * self.step
            self.reversed = not self.reversed

        if reverse:
            self.start, self.stop = self.stop, self.start
            self.step = -1 * self.step
            self.reversed = not self.reversed

    def __iter__(self):
        num = self.start
        while (not self.reversed and num <= self.stop) or (
            self.reversed and num >= self.stop
        ):
            yield self.type(num)
            num += self.step


class OptargCommand(click.Command):
    """Custom command to make the first argument optional.
    If only one argument is given, the default value for START
    is prepended to the argument list."""

    def parse_args(self, ctx, args):
        # preparse args to check if STOP was given
        parser = super(OptargCommand, self).make_parser(ctx)
        parsed, _, _ = parser.parse_args(args.copy())
        # Only START is set, prepend the default for START
        if not parsed["stop"]:
            start_arg = next(
                a for a in ctx.command.params if a.human_readable_name == "START"
            )
            args.insert(0, start_arg.default)
        super(OptargCommand, self).parse_args(ctx, args)


@click.command(cls=OptargCommand, context_settings={"ignore_unknown_options": True})
@click.option("-f", "--format", "fstring", help="formatting string for number")
@click.option("-s", "--sep", help="separator string", default="\n")
@click.option("-r", "--reverse", is_flag=True)
@click.option(
    "--also",
    multiple=True,
    nargs=3,
    help="Also run another counter with start, stop and step settings. "
    "All three need to be present.",
    metavar="START STOP STEP",
)
@click.argument("start", required=False, default="1")
@click.argument("stop")
@click.argument("step", required=False, default="1")
@click.pass_context
def run(ctx, start, stop, step, fstring, sep, reverse, also):
    """Generate a run of integers or characters. Similar to jot and seq.

    The run of numbers can be integers or reals, depending on the values of START, STOP, and STEP.
    The defaults for both START and STEP are 1. If both START and STOP are characters and STEP is an
    integer, the result will be characters."""
    fstring = interpret(fstring) or "-".join(["{}"] * (len(also) + 1))
    sep = interpret(sep) or "\n"

    try:
        counters = [Counter(start, stop, step, reverse=reverse)]
    except ValueError as err:
        ctx.fail(str(err))

    if also:
        for counter in also:
            try:
                counters.append(Counter(*counter, reverse=reverse))
            except ValueError as err:
                ctx.fail(str(err))

    run = list(product(*counters))

    # Convert to text
    try:
        runText = [fstring.format(*n) for n in run]
    except Exception as err:
        format_opt = next(
            a for a in ctx.command.params if a.human_readable_name == "fstring"
        )
        hint = format_opt.get_error_hint(ctx)
        ctx.fail(f"Invalid value for {hint}: {str(err)}")

    # Reverse the list if asked.
    # if reverse:
    #    runText.reverse()

    click.echo(sep.join(runText))



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['drang_run']

package_data = \
{'': ['*']}

install_requires = \
['click>=8.0.3,<9.0.0']

entry_points = \
{'console_scripts': ['run = drang_run:run']}

setup_kwargs = {
    'name': 'drang-run',
    'version': '0.3.5',
    'description': 'Generate a run of integers or characters. Similar to jot and seq.',
    'long_description': '=========\ndrang-run\n=========\n\n  A simple command line tool to create sequences of numbers.\n\n``drang-run`` is comparable to  ``jot`` or ``seq``, but with a more intuitive\ninterface. It was inspired (and named after) `a post by Dr. Drang\n<https://leancrew.com/all-this/2020/09/running-numbers/>`_.\n\nInstallation\n============\n\nJust install like any other package:\n\n.. code-block:: fish\n\n   pip3 install drang-run\n\nThis will install the ``run`` command.\n\n.. code-block:: fish\n\n   run --version\n\nUsage\n=====\n\nBasic usage includes up to three arguments:\n\n.. code-block:: fish\n\n   run [START] STOP [STEP]\n\n``START`` and ``STEP`` are optional and have 1 as default.\n\n.. code-block:: fish\n\n   $>run 4\n   1\n   2\n   3\n   4\n   $>run 5 8\n   5\n   6\n   7\n   8\n   $>run 0 10 3\n   0\n   3\n   6\n   9\n\nReverse the sequence with ``-r``:\n\n.. code-block:: fish\n\n   $>run 4 -r\n   4\n   3\n   2\n   1\n\nFormat the output with ``--format``. The option accepts any kind of Python format\nstring.\n\n.. code-block:: fish\n\n   $>run 998 1002 --format "{: 4}."\n    998.\n    999.\n   1000.\n   1001.\n   1002.\n\nYou can use decimals for ``START``, ``STOP`` and ``STEP``:\n\n.. code-block:: fish\n\n   $>run 1.1 1.5 .15\n   1.1\n   1.25\n   1.4\n\n.. note::\n\n   If at least one argument is a decimal, the output will be formatted as\n\t decimals as well.\n\nUsing letters will generate character sequences:\n\n.. code-block:: fish\n\n   $>run d g\n   d\n   e\n   f\n   g\n\nBy default, the sequence is separated by a newline character ``\\n``, but you can change\nthis with ``-s``:\n\n.. code-block:: fish\n\n   $>run d g -s "\\t"\n   d       e       f       g\n\nRun additional sequences with ``--also START STOP STEP``:\n\n.. code-block:: fish\n\n   $>run 1 2 -- also 3 4 1\n   1-3\n   1-4\n   2-3\n   2-4\n\nOf course, this can be used with characters and be formatted:\n\n.. code-block:: fish\n\n   $>run 1 2 -- also b c 1 --format "{0:02}. {1}_{1}"\n   01. a_a\n   01. b_b\n   02. a_a\n   02. b_b\n\n.. Note::\n\n   The sequences can be referenced by order of appearance in the format string. ``-r``\n\t will reverse *all* sequences.\n',
    'author': 'J. Neugebauer',
    'author_email': 'github@neugebauer.cc',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'http://github.com/jneug/drang-run',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6.2,<4.0.0',
}


setup(**setup_kwargs)
